from typing import Dict, List, Tuple, Any
import numpy as np
from .bfslib_if import *

# -----------------------------------------------------------------------------
#               BFSVariable
# -----------------------------------------------------------------------------
class BFSVariable:

    def __init__(self, name: str, ndims: int, datadesc: 'DataDescriptorHandle', datablock: 'DataBlockHandle'):

        self.name = name
        self._datablock = datablock
        self._datadesc = datadesc
        self.data = None
        self.axes = []
        self.target_shape = None
        self.target_type = None
        for i in range(ndims):
            status, axisname = bfs_getAxisName(datadesc, i, 1500)
            npts = bfs_getAxisNPts(datadesc, i)
            type = bfs_getAxisType(datadesc, i)
            status, units = bfs_getAxisUnits(datadesc, i, 1500)
            minval = bfs_getAxisMinValue(datadesc, i)
            maxval = bfs_getAxisMaxValue(datadesc, i)
            entry = {'axisname': axisname, 'numpoints': npts, 'bfsdatatype': type, 'units': units, 'minval': minval,
                     'maxval': maxval}
            self.axes.append(entry)

    # -----------------------------------------------------------------------------
    #               _make_shape
    # -----------------------------------------------------------------------------

    def _make_shape(self)->Tuple[List[int], int]:
        npts = 1
        shape = []
        for entry in self.axes:
            n = entry['numpoints']
            shape.append(n)
            npts *= n
        if (shape[0] == 1):
            shape = shape[1:]
        return shape, npts

    # -----------------------------------------------------------------------------
    #               axes_corder_shape
    # -----------------------------------------------------------------------------
    def axes_corder_shape(self):
        s, n = self._make_shape()
        s.reverse()
        return s

    # -----------------------------------------------------------------------------
    #               set_target_shape
    # -----------------------------------------------------------------------------
    def set_target_shape(self, newshape: List[int]):
        self.target_shape = newshape

    # -----------------------------------------------------------------------------
    #               set_target_dtype
    # -----------------------------------------------------------------------------
    def set_target_dtype(self, type: str):
        self.target_type = type

    # -----------------------------------------------------------------------------
    #               bfs_data_type
    # -----------------------------------------------------------------------------
    def bfs_data_type(self):
        return self.axes[0]['bfsdatatype']

    # -----------------------------------------------------------------------------
    #               load_data
    # -----------------------------------------------------------------------------
    def load_data(self) -> np.ndarray:

        if (self.data is None):
            shape, npts = self._make_shape()
            status, data = bfs_readParam(self._datablock, self.name, 0, npts)
            if (self.target_type is not None) or (self.target_shape is not None):
                if (self.target_type is None):
                    self.target_type = data.dtype.str
                if (self.target_shape is None):
                    shape.reverse()
                    self.target_shape = shape
                self.data = np.ndarray(shape=self.target_shape, dtype=self.target_type, buffer=data)
            else:
                shape.reverse()
                self.target_shape = shape
                self.target_type = data.dtype.str
                self.data = data.reshape(self.target_shape)




# -----------------------------------------------------------------------------
#               class BFS
# -----------------------------------------------------------------------------
class BFS:
    '''
    Python Wrapper class for the Bomem File System. It currently only supports reading BFS files.
    '''
    def __init__(self, filename: str = None):
        self._filename = 0
        self._filehandle = 0
        if (filename is not None):
            self.open(filename)

    # -----------------------------------------------------------------------------
    #               isopen
    # -----------------------------------------------------------------------------
    def isopen(self) -> bool:
        return abs(self._filehandle) > 10

    # -----------------------------------------------------------------------------
    #               close
    # -----------------------------------------------------------------------------
    def close(self):
        if self.isopen():
            bfs_close(self._filehandle)
        self._filehandle = 0

    # -----------------------------------------------------------------------------
    #                   open
    # -----------------------------------------------------------------------------
    def open(self, filename: str):
        self.close()
        self._filehandle = bfs_open(filename, "r")
        self._filename = filename

    # -----------------------------------------------------------------------------
    #               num_subfiles
    # -----------------------------------------------------------------------------
    def num_subfiles(self):
        numsubfiles = bfs_getNumberOfSubfiles(self._filehandle) if self.isopen() else 0
        return numsubfiles

    # -----------------------------------------------------------------------------
    #               _load_parameter
    # -----------------------------------------------------------------------------
    def _load_parameter(self, name, npts, type, datablock) -> Any:

        value = None                                                        # default value is to return None
        if (type == TYPE_ASCII_STRING) or (type == TYPE_STRING):            # see if we have a string
            status, value = bfs_readString(datablock, name, npts)           # And get the string
        else:                                                               # otherwise load in the numeric values
            status, value = bfs_readParam(datablock, name, 0, npts)         # Returns None or an np.ndarray
            if (value is None):                                             # Check for errors
                print('Error looking up parameter {}'.format(name), ', npts = ', npts, 'type =', type)
            else:                                                           # and if we have just one point
                if (npts == 1):                                             # then convert it into
                    value = value[0]                                        # a scalar value.
        return value

    # -----------------------------------------------------------------------------
    #                   load_parameters_from_header
    # -----------------------------------------------------------------------------
    def load_header(self) -> Dict[str, Any]:

        parameters = {}
        if self.isopen():
            header = bfs_getHeaderDataBlock(self._filehandle)
            numparam = bfs_getDataDescriptorNbrEntries(header)
            for index in range(numparam):
                datadesc = bfs_getDataDescriptorByIndex(header, index)      # Get the BFS data descriptor
                status, name = bfs_getName(datadesc, 1500)                  # Get the name of this parameter
                npts = bfs_getAxisNPts(datadesc,0)                          # Get the number of points in this parameter. The first field of the axis arrays
                type = bfs_getAxisType(datadesc, 0)                         # Get the data type of this parameter
                ok = (name is not None) and (len(name) > 0)                 # If we have a none blank name
                if (ok):  # then
                    value = self._load_parameter(name, npts, type, header)
                    parameters[name] = value
        return parameters

    # -----------------------------------------------------------------------------
    #               load_datafields
    # -----------------------------------------------------------------------------
    def load_datafields(self, subfileindex: int = 0, loaddata: bool= True) -> Tuple[Dict[str, BFSVariable], Dict[str, Any]]:
        fields = {}
        parameters = {}
        if self.isopen():
            bfs_readSubfile(self._filehandle, subfileindex)
            datablock = bfs_getSubfileDataBlock(self._filehandle)
            numparam = bfs_getDataDescriptorNbrEntries(datablock)
            for index in range(numparam):
                datadesc = bfs_getDataDescriptorByIndex(datablock, index)   # Get the BFS data descriptor
                status, name = bfs_getName(datadesc, 1500)                  # Get the name of this parameter ndims = bfs_getNbrDims(datadesc)
                ok = (name is not None) and (len(name) > 0)
                if (ok):
                    ndims = bfs_getNbrDims(datadesc)
                    if ndims == 1:
                        npts = bfs_getAxisNPts(datadesc, 0)
                        type = bfs_getAxisType(datadesc, 0)
                        value = self._load_parameter(name, npts, type, datablock)
                        parameters[name] = value
                    else:
                        value = BFSVariable(name, ndims, datadesc, datablock)
                        if loaddata:
                            value.load_data()
                        fields[name] = value
        return fields, parameters
